import asyncio
import ctypes
import json
import logging
import time
from binascii import hexlify
from collections import defaultdict

import zmq.asyncio
from distutils.version import LooseVersion
from google.protobuf import json_format
from keroslib import utils
from loraapi import common_lora_pb2
from loraapi import control_channel_pb2
# Set of modules only available since keros 5.10
try:
    from loraapi import control_beacon_pb2
    from loraapi import control_freqrange_pb2
    from loraapi import control_lbt_pb2
except ImportError:
    pass
from loraapi import control_rpc_pb2
from loraapi import control_version_pb2
from loraapi import downlink_container_pb2
from loraapi import downlink_gfsk_pb2
from loraapi import downlink_lora_pb2
from loraapi import uplink_container_pb2
from . import lorawan

log = logging.getLogger('lorad')


def build_lora_downlink_message(msgid, payload, txconfig):
    SFs = {7: common_lora_pb2.SF_7,
           8: common_lora_pb2.SF_8,
           9: common_lora_pb2.SF_9,
           10: common_lora_pb2.SF_10,
           11: common_lora_pb2.SF_11,
           12: common_lora_pb2.SF_12
           }
    BWs = {125: common_lora_pb2.BW_125K,
           250: common_lora_pb2.BW_250K,
           500: common_lora_pb2.BW_500K
           }

    msg = downlink_lora_pb2.Downlink()
    msg.message_id = msgid
    msg.payload = payload
    msg.required_transmissions = 1
    # LoRaWAN downlink defaults
    msg.polarity_inverted = True
    msg.crc_enabled = False
    msg.header_disabled = False
    # Transmissions
    configuration = msg.configurations.add()
    configuration.radio_id = txconfig['rid']
    configuration.frequency = txconfig['frq']
    configuration.bandwidth = BWs[txconfig['bw']]
    configuration.spreading_factor = SFs[txconfig['sf']]
    configuration.power = txconfig['pwr']
    timing = configuration.timings.add()
    timing.transmission_id = 1
    timing.priority = txconfig.get('prio', 0xFFFFFFFF)

    if 'gpstime' in txconfig:
        timing.time.time = int(txconfig['gpstime']/1000000)
        timing.time.microseconds = txconfig['gpstime'] % 1000000
        # "GPS_EPOCH_TAI_TIME" from
        # common_time_pb2.AccurateContinuousTime.ContinuousTimeFormat
        timing.time.format = 1

    # do not set counter field for immediate downlinks
    if txconfig['cnt'] is not None:
        timing.counter = ctypes.c_uint32(txconfig['cnt']).value

    return msg


def build_gfsk_downlink_message(msgid, payload, txconfig):
    msg = downlink_gfsk_pb2.Downlink()
    msg.message_id = msgid
    msg.payload = payload
    msg.required_transmissions = 1
    # LoRaWAN downlink defaults
    msg.length_fixed = False
    msg.crc = False
    msg.encoding = False
    # Transmissions
    configuration = msg.configurations.add()
    configuration.radio_id = txconfig['rid']
    configuration.frequency = txconfig['frq']
    configuration.bitrate = 50000
    configuration.power = txconfig['pwr']
    timing = configuration.timings.add()
    timing.transmission_id = 1
    timing.priority = txconfig.get('prio', 0xFFFFFFFF)
    timing.counter = txconfig['cnt']

    return msg


def compute_frequency(radios, channel):
    try:
        if channel['enable'] and radios[channel['radio']]['enable']:
            return radios[channel['radio']]['frequency'] + channel['if']
        else:
            log.warning(
                f"A parameter may be uncorrect to compute channel \
                frequency: radio list {radios} vs channel {channel}")
            return 0
    except Exception:
        log.warning(
            f"A parameter may be uncorrect to compute channel \
            frequency: radio list {radios} vs channel {channel}")
        return 0


def get_max_tx_power(region: str, freq: int, jplms=False):
    """Returns max TX power (dBm EIRP) for a given region and frequency

        Parameters
        ----------
        region: str
            String describing region, ex: 'EU868' or 'US915'

        freq: int
            TX frequency (in Hz), ex: 869525000

        jplms: bool, optionnal
            Use Japan LMS standard (default = False)

        Returns
        -------
        tx_power: int
            Maximum EIRP power to use (in dBm)
    """
    DEFAULT_TX_POWER = 14

    if region in ['EU863', 'EU868']:
        # 16 dbm excepting RX2 / multicast - 869.525 - 29dBM
        if freq == 869525000:
            tx_power = 29
        else:
            tx_power = 16

    elif region in ['US902', 'US915', 'AU915', 'IN865']:
        # 36 dbm on whole band
        tx_power = 36

    elif region in ['AS923', 'AS923JP']:
        # 920.5 to 923.5 MHz
        if jplms and 920500000 <= freq <= 923500000:
            tx_power = 27
        else:
            # Highly depending on country ... set 16dBm by default
            tx_power = 16

    else:
        # Unknown region, set default power
        tx_power = DEFAULT_TX_POWER

    return tx_power


class Lorad():
    def __init__(self, uplink, downlink, control):
        ctx = zmq.asyncio.Context()
        sockup = ctx.socket(zmq.SUB)
        sockup.connect(uplink)
        sockup.setsockopt(zmq.SUBSCRIBE, b"")
        self.sockup = sockup

        sockdw = ctx.socket(zmq.PUSH)
        sockdw.connect(downlink)
        self.sockdw = sockdw

        self.control = control
        self.ctx = ctx
        self.renew_control_socket()
        self.ctrl_tmo = 10
        self.max_retries = 5

        self.rx = {
            "lora": asyncio.Queue(),
            "gfsk": asyncio.Queue()
        }
        self.tx_waiter = defaultdict(list)

        # some APIs are only available in Keros 5.10
        self.extra_apis_available = False

    def renew_control_socket(self):
        sockct = self.ctx.socket(zmq.REQ)
        sockct.connect(self.control)
        self.sockct = sockct

    async def apply_radio_conf(self, region, radio_conf):
        '''Apply radio configuration is compatible with platform'''

        SFs = {7: common_lora_pb2.SF_7,
               8: common_lora_pb2.SF_8,
               9: common_lora_pb2.SF_9,
               10: common_lora_pb2.SF_10,
               11: common_lora_pb2.SF_11,
               12: common_lora_pb2.SF_12}
        BWs = {125000: common_lora_pb2.BW_125K,
               250000: common_lora_pb2.BW_250K,
               500000: common_lora_pb2.BW_500K}

        # Retrieve platform region and check compatibility with expected region
        if "IBTS" not in utils.getPlatform():
            platform_region = utils.getRegion().rstrip()
            # Match only on the first 3 characters as US915, for instance
            # can also be refered to as US902 depending on the type of LNS
            if not ((platform_region == '868' and region[:3] in
                    ('EU8', 'IN8')) or
                    (platform_region == '915' and region[:3] in
                    ('US9', 'AU9')) or
                    (platform_region == '923' and region[:3] in
                    ('AS9', 'AU9'))):
                log.debug(
                    f"Region {region} not compatible with \
                  platform region {platform_region}")

        log.debug("Platform Name : " + utils.getPlatform())

        # Convert radio_conf to lorad API

        # Create protobuf request
        req = control_channel_pb2.Request()

        # Create radio list
        radios = {}
        for radio_id in range(2):
            try:
                radios[radio_id] = \
                    {'enable': radio_conf['radio_' + str(radio_id)]['enable'],
                     'frequency': radio_conf['radio_' + str(radio_id)]['freq']}

                if self.extra_apis_available is True:
                    freq = int(radios[radio_id]['frequency'])
                    req.radio_frequencies.append(freq)

            except Exception:
                log.info(f"No radio {radio_id} parameters found")

        channel_number = 0

        # Multi SF channel
        try:
            for channel_id in range(8):
                frequency = compute_frequency(
                    radios, radio_conf['chan_multiSF_' + str(channel_id)])
                if frequency:
                    channel = req.channels.add()
                    channel.modulation = control_channel_pb2.LoRa
                    channel.central_frequency = frequency
                    channel.bandwidth = BWs[125000]
                    data_rate = channel.data_rates.add()
                    data_rate.spreading_factor = SFs[7]
                    data_rate = channel.data_rates.add()
                    data_rate.spreading_factor = SFs[8]
                    data_rate = channel.data_rates.add()
                    data_rate.spreading_factor = SFs[9]
                    data_rate = channel.data_rates.add()
                    data_rate.spreading_factor = SFs[10]
                    data_rate = channel.data_rates.add()
                    data_rate.spreading_factor = SFs[11]
                    data_rate = channel.data_rates.add()
                    data_rate.spreading_factor = SFs[12]

                    channel_number += 1
        except Exception:
            log.info(
                f"Uncorrect set of parameters for channel \
                chan_multiSF_{channel_id}")

        # Mono SF channel
        try:
            frequency = compute_frequency(radios, radio_conf['chan_Lora_std'])
            if frequency:
                channel = req.channels.add()
                channel.modulation = control_channel_pb2.LoRa
                channel.central_frequency = frequency
                channel.bandwidth = BWs[radio_conf['chan_Lora_std']
                                        ['bandwidth']]
                data_rate = channel.data_rates.add()
                data_rate.spreading_factor = SFs[radio_conf['chan_Lora_std']
                                                 ['spread_factor']]

                channel_number += 1
        except Exception:
            log.info("Uncorrect set of parameters for channel chan_Lora_std")

        # FSK channel
        try:
            frequency = compute_frequency(radios, radio_conf['chan_FSK'])
            if frequency:
                channel = req.channels.add()
                channel.modulation = control_channel_pb2.GFSK
                channel.central_frequency = frequency
                channel.bandwidth = BWs[125000]
                data_rate = channel.data_rates.add()
                data_rate.bitrate = 50000

                channel_number += 1
        except Exception:
            log.info("Uncorrect set of parameters for channel chan_FSK")

        # Check if at least a channel has been found
        if channel_number:
            log.debug(f"Going to set {channel_number} radio channels")
            # Call lorad API
            # Warning : lorad may restart to apply new conf, and not respond
            attempt = 1
            while (attempt < self.max_retries):
                try:
                    protobuf = await asyncio.wait_for(self.rpc(
                        "set_channels", req.SerializeToString()),
                        timeout=self.ctrl_tmo)
                    # debug log to 'use' above variable (flake8)
                    log.debug("Set Channels : " + str(protobuf))
                    attempt = self.max_retries
                except asyncio.TimeoutError:
                    log.warning(f"no RPC response from lorad (try #{attempt})")
                    attempt = attempt + 1
                    self.renew_control_socket()
        else:
            log.warning(
                "Invalid radio configuration, no valid channel to configure")

        # Configure Listen Before Talk if provided
        if 'lbt_cfg' in radio_conf:
            await self.apply_lbt_config(radio_conf['lbt_cfg'])
        else:
            await self.apply_lbt_config({'enable': False, 'rssi_target': -80})

        return

    async def apply_lbt_config(self, lbt_config):
        if self.extra_apis_available is False:
            log.debug("LBT Configuration API not available in this version.")
            return
        request = control_lbt_pb2.Request()

        try:
            request.lbt_enable = lbt_config['enable']
            request.lbt_rssi_target = lbt_config['rssi_target']
            i = 0
            if lbt_config.get('chan_cfg') is not None:
                for chan_cfg in lbt_config['chan_cfg']:
                    channel_config = request.channels_config.add()
                    channel_config.freq_hz = chan_cfg['freq_hz']
                    channel_config.scan_time_us = chan_cfg['scan_time_us']
                    if chan_cfg.get('transmit_time_ms') is not None:
                        channel_config.transmit_time_ms = \
                            chan_cfg['transmit_time_ms']
                    i = i + 1
        except Exception as e:
            log.error(f"Unexpected JSON file content: {e}")
            raise SyntaxError('Invalid lbt_cfg syntax')

        attempt = 1
        while (attempt < self.max_retries):
            try:
                protobuf = await asyncio.wait_for(self.rpc(
                    "set_lbt", request.SerializeToString()),
                    timeout=self.ctrl_tmo)
                # debug log to 'use' above variable (flake8)
                log.debug("Set LBT : " + str(protobuf))
                attempt = self.max_retries
            except asyncio.TimeoutError:
                log.warning(f"no RPC response from lorad (try #{attempt})")
                attempt = attempt + 1
                self.renew_control_socket()

        return

    async def apply_freq_range(self, freq_ranges):
        if self.extra_apis_available is False:
            log.debug("Freq Range API not available in this version.")
            return
        '''Apply TX Frequency Range'''
        req = control_freqrange_pb2.Request()
        freqrange = req.freqranges.add()
        freqrange.tx_freq_min = freq_ranges[0]
        freqrange.tx_freq_max = freq_ranges[1]
        attempt = 1
        while (attempt < self.max_retries):
            try:
                protobuf = await asyncio.wait_for(self.rpc(
                    "set_freqranges", req.SerializeToString()),
                    timeout=self.ctrl_tmo)
                # debug log to 'use' above variable (flake8)
                log.debug("Set Freq Range : " + str(protobuf))
                attempt = self.max_retries
            except asyncio.TimeoutError:
                log.warning(f"no RPC response from lorad (try #{attempt})")
                attempt = attempt + 1
                self.renew_control_socket()

        return

    async def apply_beaconing_conf(self, bcning, region):
        if self.extra_apis_available is False:
            log.debug("Beaconing API not available in this version.")
            return
        '''Apply Beacon Settings'''
        req = control_beacon_pb2.Request()
        if bcning is not None:
            _DR_ = bcning['DR']
            _FREQS_ = bcning['freqs']
            _EN_ = True
        else:
            _DR_ = 0
            _FREQS_ = [0]
            _EN_ = False
        log.debug(f" beacon reg : {region} , DR : {_DR_}")
        try:
            sf, bw_index = lorawan.DR_to_SF[region][_DR_]
        except KeyError:
            log.warning(f"Beacon region {region} not handled.")
            return
        log.debug(f" => SF : {sf} , BW : {lorawan.BWs[bw_index]}")
        bandwidth = lorawan.BWs[bw_index]
        req.beacon_enable = _EN_
        freqs = _FREQS_
        freqs.sort()
        req.beacon_freq_nb = len(freqs)
        req.beacon_freq_hz = freqs[0]
        req.beacon_sf = sf
        req.beacon_bw_hz = bandwidth

        if len(freqs) > 1:
            req.beacon_freq_step = freqs[1] - freqs[0]
        else:
            req.beacon_freq_step = 0

        attempt = 1
        while (attempt < self.max_retries):
            try:
                protobuf = await asyncio.wait_for(self.rpc(
                    "set_beacons", req.SerializeToString()),
                    timeout=self.ctrl_tmo)
                # debug log to 'use' above variable (flake8)
                log.debug("Set Beacons : " + str(protobuf))
                attempt = self.max_retries
            except asyncio.TimeoutError:
                log.warning(f"no RPC response from lorad (try #{attempt})")
                attempt = attempt + 1
                self.renew_control_socket()

        return

    async def apply_changes(self):
        if self.extra_apis_available is False:
            log.debug("Apply Configuration API not available in this version.")
            return

        attempt = 1
        while (attempt < self.max_retries):
            try:
                protobuf = await asyncio.wait_for(self.rpc(
                    "apply_configuration", None),
                    timeout=self.ctrl_tmo)
                log.debug("Apply Changes : " + str(protobuf))
                attempt = self.max_retries
            except asyncio.TimeoutError:
                log.warning(f"no RPC response from lorad (try #{attempt})")
                attempt = attempt + 1
                self.renew_control_socket()

    async def recv_task(self):
        '''Task listening to the uplink socket and doing uplink triage'''
        while True:
            data = await self.sockup.recv()
            log.debug("Lorad: received data " + hexlify(data).decode("ascii"))
            c = uplink_container_pb2.Container()
            c.ParseFromString(data)
            msg = json_format.MessageToDict(c)
            msgtype = [*msg][0]
            log.info(f'"{msgtype}" message received')
            log.debug(f'{msg}')
            if msgtype == "event":
                # Put msg in tx_waiter queue(s) if any
                # (tasks are waiting for that event)
                msgid = c.event.transmission.message_id
                if msgid in self.tx_waiter:
                    await asyncio.gather(*[q.put(msg)
                                         for q in self.tx_waiter[msgid]])
            elif msgtype in ("lora", "gfsk"):
                # Put msg in queue, use recv() to get it
                await self.rx[msgtype].put(msg)
            else:
                # TODO Handle 'scan' messages
                log.warning(f"Unhandled message type \"{msgtype}\"")

    async def recv(self, msgtype):
        return await self.rx[msgtype].get()

    async def send(self, msgid, payload, txconfig):
        container = downlink_container_pb2.Container()

        if txconfig['sf'] in range(7, 13):
            container.lora.CopyFrom(
                build_lora_downlink_message(msgid, payload, txconfig))
        elif txconfig['sf'] == 0:
            container.gfsk.CopyFrom(
                build_gfsk_downlink_message(msgid, payload, txconfig))
        else:
            log.warning(
                "Unexpected message type with SF: {}".format(txconfig['sf']))
            return

        log.debug(f"Sending data: {json_format.MessageToDict(container)}")

        await self.sockdw.send(container.SerializeToString())
        return json_format.MessageToDict(container)

    async def wait_txevent(self, msgid):
        '''Register and wait on `msgid` transmission event'''
        q = asyncio.Queue()
        self.tx_waiter[msgid].append(q)
        msg = await q.get()
        log.debug(f"Event on Queue[{msgid}] received! {msg}")
        msg["event"]["transmission"]["txtime"] = time.time()
        # Clean-up
        self.tx_waiter[msgid].remove(q)
        if not self.tx_waiter[msgid]:
            del self.tx_waiter[msgid]
        return msg["event"]

    async def rpc(self, service, args):
        req = control_rpc_pb2.Request()
        req.service = service
        if args is not None:
            req.protobuf = args
        await self.sockct.send(req.SerializeToString())
        protobuf = await self.sockct.recv()
        reply = control_rpc_pb2.Reply()
        reply.ParseFromString(protobuf)
        if reply.status == control_rpc_pb2.Reply.FAILED:
            raise RuntimeError("The RPC request failed")
        if reply.status == control_rpc_pb2.Reply.BAD_REQUEST:
            raise RuntimeError("The RPC request is incorrect")
        if reply.status == control_rpc_pb2.Reply.NOT_IMPLEMENTED:
            log.warning("The " + service + " RPC service is not implemented")
        return reply.protobuf

    async def check_lorad_version(self):
        protobuf = ""
        attempt = 1
        while (attempt < self.max_retries):
            try:
                protobuf = await asyncio.wait_for(self.rpc(
                    "get_version", None), timeout=3 * self.ctrl_tmo)
                attempt = self.max_retries
            except TimeoutError:
                log.warning(f"no RPC response from lorad (try #{attempt})")
                self.renew_control_socket()
                if attempt >= self.max_retries:
                    raise RuntimeError(
                        f"No response from lorad within\
                        {attempt * 3 * self.ctrl_tmo} seconds")
                attempt = attempt + 1

        reply = control_version_pb2.Reply()
        reply.ParseFromString(protobuf)
        rep_json = json.loads(json_format.MessageToJson(reply))
        log.info(rep_json)

        try:
            version = rep_json['version'].split("lorad ")[1]
            if LooseVersion(version) >= LooseVersion("2.2.0"):
                log.debug("latest APIs are available")
                self.extra_apis_available = True
        except KeyError:
            raise RuntimeError("Failed to parse lorad version")

        return

    async def versions(self):
        # check lorad version to establish the
        # availability of latest APIs
        await self.check_lorad_version()

        # replace the original statistics fpga + hal + dsp retrieval
        # through protobuf with a more concise keros version info as
        # it provides the same level of detail in the end.
        res = "unknown"
        with open('/etc/version') as verfile:
            res = "Keros-" + verfile.read().strip()
        return res
