import asyncio
import ctypes
import json
import logging
import os
import random
import socket
import ssl
import struct
import subprocess
import sys
from binascii import hexlify
from math import log10
from socket import gaierror
from urllib.parse import urlparse

import websockets

from .cidr import Set
from .config import backup
from .cups import connection_args
from .cups import handle_firewall
from .id6 import Eui
from .id6 import Id6

if sys.version_info[1] < 7:
    asyncio.create_task = asyncio.ensure_future

log = logging.getLogger('router')

deco_trigger = os.environ.get('DECO_TRIGGER')


class RouterClient():
    '''Basic Station LNS client (packet forwarder)'''

    stats_up_msg_recv = 0
    stats_up_msg_sent = 0
    stats_join_req_recv = 0
    stats_join_req_sent = 0
    stats_dn_msg_recv = 0
    stats_dn_rx1 = 0
    stats_dn_rx2 = 0
    stats_devaddr_filter = []
    stats_joineui_filter = []

    def __init__(self, rtcfg):
        self.routerid = Id6(rtcfg['id'])
        self.connected = asyncio.Event()
        self.router_config = asyncio.Event()
        self.rxq = asyncio.Queue()
        self.versions = {}
        self.jplms = rtcfg['jplms']
        self.ping_int = rtcfg['ping_interval']
        self.local_filters = rtcfg['filters']
        self.setLed(False, True)
        self.show_delayed = rtcfg['show_delayed']
        self.sock_tmo = rtcfg['sock_tmo']

    def getStats(self):
        retVal = dict()
        retVal["lorawan_uplink_msg_recv"] = self.stats_up_msg_recv
        retVal["lorawan_uplink_msg_sent"] = self.stats_up_msg_sent
        retVal["join_req_recv"] = self.stats_join_req_recv
        retVal["join_req_sent"] = self.stats_join_req_sent
        retVal["downlink_msg_recv"] = self.stats_dn_msg_recv
        retVal["downlink_msg_RX1_conf"] = self.stats_dn_rx1
        retVal["downlink_msg_RX2_conf"] = self.stats_dn_rx2

        retVal["devaddr_filter"] = self.stats_devaddr_filter
        retVal["join_eui_filter"] = self.stats_joineui_filter

        return retVal

    def setLed(self, green, red):
        GreenLed = "/sys/class/leds/led2:green:backhaul/brightness"
        RedLed = "/sys/class/leds/led2:red:backhaul/brightness"

        if (not os.path.exists(GreenLed)) or (not os.path.exists(RedLed)):
            return

        if green is True:
            subprocess.call("echo 1 > " + GreenLed, shell=True)
            subprocess.call("echo 0 > " + RedLed, shell=True)
        elif red is True:
            subprocess.call("echo 0 > " + GreenLed, shell=True)
            subprocess.call("echo 1 > " + RedLed, shell=True)
        else:
            subprocess.call("echo 0 > " + GreenLed, shell=True)
            subprocess.call("echo 0 > " + RedLed, shell=True)

    async def connect(self, tcuri, tctrust=None,
                      tccrt=None, tckey=None, capath=None):
        '''Connect to INFOS (LNS Discovery) to get MUXS URI
           Connect to MUXS  (LNS Data) if all went well'''

        self.tcuri = tcuri
        if not self.tcuri:
            log.critical("No TC URI found: abort")
            raise ValueError("No TC URI")

        self.trust = tctrust
        self.key = tckey
        self.crt = tccrt
        self.capath = capath

        await self._connect()

    async def _connect(self):
        global deco_trigger
        log.debug("Connecting to {tcuri}, with credentials: {trust},\
            {crt}, {key}".format(
            **self.__dict__))
        # used for dev/debug purposes only
        if (deco_trigger is not None) and (os.path.isfile(deco_trigger)):
            return

        if self.tcuri is None:
            # tcuri is probably being updated by CUPS ..
            log.debug("Awaiting new TC URI")
            return

        uri = self.tcuri + "/router-info"
        msg = {"router": str(self.routerid)}
        # TLS and Authentication
        wsargs = connection_args(self.trust, self.crt,
                                 self.key, self.capath, 'extra_headers')

        try:
            handle_firewall(uri)
            async with websockets.connect(uri, ping_interval=None,
                                          ** wsargs) as ws:
                await ws.send(json.dumps(msg))
                rsp = await ws.recv()
            router_info = json.loads(rsp)
            if "error" in router_info:
                log.error("router-info error: {error}".format(**router_info))
                return
            self.mux = router_info["muxs"]
            self.muxuri = router_info["uri"]
            log.info("router-info MUX {muxs} @ {uri}".format(**router_info))
            handle_firewall(self.muxuri)

            uri_info = urlparse(self.muxuri)
            mux_socket = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
            mux_socket.setsockopt(socket.IPPROTO_TCP, socket.TCP_USER_TIMEOUT,
                                  self.sock_tmo * 1000)
            srv_hostname = None
            if uri_info.scheme == 'wss':
                # this field shall only be provided when SSL is involved
                srv_hostname = uri_info.hostname
            uri_port = uri_info.port
            if uri_port is None:
                # affect default ports values if not provided by server
                if uri_info.scheme == 'wss':
                    uri_port = 443
                else:
                    uri_port = 80
            loop = asyncio.get_event_loop()
            await loop.sock_connect(mux_socket,
                                    (uri_info.hostname, uri_port))

            self.ws = \
                await websockets.connect(self.muxuri, sock=mux_socket,
                                         server_hostname=srv_hostname,
                                         ping_interval=self.ping_int,
                                         ** wsargs)
            self.connected.set()
            self.setLed(True, False)
            log.info("Connected to LNS " + self.muxuri)
            log.debug("Websockets Ping Interval set to " + str(self.ping_int))
            # router-info transaction was successfull,
            # back up the TC credentials
            backup("tc", "-bak")

        except (ConnectionRefusedError, TimeoutError, OSError,
                websockets.exceptions.WebSocketException,
                gaierror, ssl.SSLError, asyncio.exceptions.TimeoutError) as e:
            log.debug("ws connection error : " + str(e))
            await asyncio.sleep(1)

    async def disconnect(self):
        if self.connected.is_set():
            self.connected.clear()
            self.router_config.clear()
            log.info("Disconnected from LNS " + self.muxuri)
            self.setLed(False, True)
            await self.ws.close()

    async def _send(self, data):
        global deco_trigger
        '''Internal method to send only when connected'''
        await self.connected.wait()
        # used for dev/debug purposes only
        if (deco_trigger is not None) and (os.path.isfile(deco_trigger)):
            log.warning("FILE TRIGGERED DISCONNECTION")
            await self.disconnect()
        try:
            await self.ws.send(data)
        except websockets.exceptions.WebSocketException as e:
            log.warning("send failure : " + str(e))
            await self.disconnect()

    async def _recv(self):
        '''Internal method to recv only when connected'''
        await self.connected.wait()
        return await self.ws.recv()

    async def send_version(self, versions={}):
        '''Connect to MUXS and send "version" message'''
        # keep track of versions in case of reconnection
        self.versions = versions
        msg = {
            "msgtype": "version",
            "station": versions.get('station'),
            "firmware": versions.get('firmware'),
            "package": versions.get('package'),
            "model": versions.get('model'),
            "protocol": 2,
            "features": "prod"
        }
        log.debug(f"Sending version: {msg}")
        await self._send(json.dumps(msg))

    async def uplink(self, loramsg, freq, sf, bw, rctx, xtime, rssi,
                     snr=None, rcvtime=None, delayed=False):
        '''We got something to send to MUXS: sort it out and send'''
        try:
            dr = self.DRs.index([sf, bw, 0])
        except ValueError:
            log.warning(f"Unknown Datarate ({sf},{bw}): discard uplink")
            return
        except AttributeError:
            log.warning("Datarate table not ready: discard uplink")
            return

        radio_metadata = {
            "DR": dr,
            "Freq": freq,
            "upinfo": {
                "rctx": rctx,
                "xtime": xtime,
                "rssi": rssi,
            }
        }
        if snr is not None:
            radio_metadata['upinfo']['snr'] = snr
        if not loramsg:
            log.warning("Empty uplink message: discard")
            return

        if rcvtime is not None:
            radio_metadata['upinfo']['gpstime'] = rcvtime

        # Check MHDR for Join Request or other uplink
        (mhdr,) = struct.unpack_from("<B", loramsg)
        ftype = (mhdr & 0xE0) >> 5
        if ftype == 0 and len(loramsg) == 23:  # Join Request
            log.info("LoRaWAN \"Join-Request\" message")
            self.stats_join_req_recv = self.stats_join_req_recv + 1
            await self.jreq(loramsg, radio_metadata)
        # {un}confirmed data uplink
        elif ftype in [2, 4] and len(loramsg) >= 12:
            log.info("LoRaWAN \"{}confirmed data uplink\" message".format(
                {2: "un", 4: ""}[ftype]))
            self.stats_up_msg_recv = self.stats_up_msg_recv + 1
            await self.updf(loramsg, radio_metadata, rcvtime, delayed)
        elif ftype == 0b111:
            log.info("LoRaWAN \"Proprietary\" message")
            await self.propdf(loramsg, radio_metadata)
        else:
            log.warning(
                f"Non-LoRaWAN uplink message (type {ftype:#b},\
                len {len(loramsg)}): discard")

    async def updf(self, loramsg, radio_metadata, rcvtime=None, delayed=False):
        msg = {"msgtype": "updf"}
        msg.update(radio_metadata)

        keys = ("MHdr", "DevAddr", "FCtrl", "FCnt")
        vals = struct.unpack_from("<BiBH", loramsg)
        msg.update(zip(keys, vals))
        # filter on DevAddr
        if (self.DevAddrRanges is not None) and (len(self.DevAddrRanges) > 0):
            allowed = False
            devAddr = ctypes.c_uint32(msg['DevAddr']).value
            for Range in self.DevAddrRanges:
                if (devAddr >= Range[0]) and \
                   (devAddr <= Range[1]):
                    allowed = True
                    log.info(hex(devAddr) + " FILTER OK ")
                    break

            if allowed is False:
                log.info(hex(devAddr) + " FILTERED OUT ")
                return

        fopts_len = msg["FCtrl"] & 0x0F
        msg["FOpts"] = hexlify(loramsg[8:8+fopts_len]).decode("ascii")
        nextbyte = 8+fopts_len

        if loramsg[nextbyte:-4]:
            msg["FPort"], = struct.unpack_from("<B", loramsg, nextbyte)
            msg["FRMPayload"] = hexlify(loramsg[nextbyte+1:-4]).decode("ascii")
        else:
            msg["FPort"] = -1
            msg["FRMPayload"] = ""

        msg["MIC"], = struct.unpack_from("<i", loramsg, -4)

        if delayed is True:
            if self.show_delayed is True:
                msg["delayed"] = 1
                # one day we may want to also store the timestamp
                # the message has been bufferized at ..
                # msg["buffer_time"] = rcvtime

        log.debug(f"Sending {msg}")
        self.stats_up_msg_sent = self.stats_up_msg_sent + 1
        await self._send(json.dumps(msg))

    async def jreq(self, loramsg, radio_metadata):
        msg = {"msgtype": "jreq"}
        msg.update(radio_metadata)

        keys = ("MHdr", "JoinEui", "DevEui", "DevNonce", "MIC")
        vals = struct.unpack_from("<BQQHi", loramsg)
        msg.update(zip(keys, vals))
        __join_eui__ = int(msg["JoinEui"])
        for eui in "JoinEui", "DevEui":
            msg[eui] = Eui(msg[eui]).euistr
        # filter on JoinEui
        if (self.JoinEui is not None) and (len(self.JoinEui) > 0):
            allowed = False
            for JOIN_EUI in self.JoinEui:
                if (__join_eui__ >= JOIN_EUI[0]) and \
                   (__join_eui__ <= JOIN_EUI[1]):
                    allowed = True
                    log.info(hex(__join_eui__)[2:] + " JOIN EUI OK")
                    break

            if allowed is False:
                log.info(hex(__join_eui__)[2:] + " JOIN EUI FILTERED OUT ")
                return

        log.debug(f"Sending {msg}")
        self.stats_join_req_sent = self.stats_join_req_sent + 1
        await self._send(json.dumps(msg))

    async def propdf(self, loramsg, radio_metadata):
        msg = {"msgtype": "propdf"}
        msg.update(radio_metadata)
        msg["FRMPayload"] = hexlify(loramsg).decode("ascii")
        self.stats_up_msg_sent = self.stats_up_msg_sent + 1
        await self._send(json.dumps(msg))

    async def dntxed(self, diid, deveui, rctx, xtime, txtime):
        msg = {
            "msgtype": "dntxed",
            "diid": diid,
            "DevEui": Eui(deveui).euistr,
            "rctx": rctx,
            "xtime": xtime,
            "txtime": txtime,
            "gpstime": 0
        }
        log.debug(f"Sending {msg}")
        self.stats_up_msg_sent = self.stats_up_msg_sent + 1
        await self._send(json.dumps(msg))

    def computeDevAddrRanges(self):
        nwkIdLen = {'0': 6, '1': 6, '2': 9, '3': 11,
                    '4': 12, '5': 13, '6': 15, '7': 17}
        ADDR_RANGES = []
        self.stats_devaddr_filter = []

        if (self.NetID is None) and (self.local_filters is not None):
            log.debug("Applying local filters " + self.local_filters)
            S = Set()
            S.load_cidr_intervals(self.local_filters.split(","))
            ADDR_RANGES = S.get_intervals()

        if (self.NetID is not None) and (len(self.NetID) > 0):
            for NET_ID in self.NetID:
                # Retrieve type from Net ID
                NetIdTypeStr = bin(NET_ID)[2:].zfill(24)[:3]
                NetIdTypeInt = int(NetIdTypeStr, 2)

                # Compute DevAddr type based prefix
                PREFIX = NetIdTypeInt*'1' + '0'

                # Retrieve NwkID size from table
                nwkidlen = nwkIdLen[str(NetIdTypeInt)]

                # Work out NwkAddr size from other fields lengths
                NwkAddrLen = 32 - (nwkidlen + len(PREFIX))

                # Assemble binary value of the first available DevAddr
                firstAddrBin = PREFIX
                firstAddrBin += bin(NET_ID)[2:].zfill(nwkidlen)[-nwkidlen:]
                firstAddrBin += NwkAddrLen*'0'

                # Work out address range limits
                first = int(firstAddrBin, 2)
                last = first + int(NwkAddrLen * '1', 2)
                ADDR_RANGES.append([first, last])
        for [FIRST_ADDR, LAST_ADDR] in ADDR_RANGES:
            log.info("DevAddr allowed from " +
                     hex(FIRST_ADDR) + " to " + hex(LAST_ADDR))
            self.stats_devaddr_filter.append([f'{FIRST_ADDR:x}',
                                              f'{LAST_ADDR:x}'])
        self.DevAddrRanges = ADDR_RANGES

    async def recv_task(self):
        '''Task listening to the MUX websocket and doing msg triage'''
        DELAY_INIT = 1
        MAX_DELAY = 600
        delay = DELAY_INIT
        cumulated_deco_time = 0

        # do not attempt a 'reconnect' if no tc.uri has been retrieved yet
        tcuri_wait = 0
        while not hasattr(self, 'tcuri'):
            log.debug("tcuri not configured yet ..")
            await asyncio.sleep(1)
            tcuri_wait = tcuri_wait + 1
            if tcuri_wait > 120:
                raise RuntimeError("Unable to retrieve TC.URI")

        while True:
            if not self.connected.is_set():
                log.warning("reconnection attempt ..")
                await self._connect()
                # if reconnection has succeeded, send version
                if self.connected.is_set():
                    await self.send_version(self.versions)
                else:
                    log.debug("next attempt in " + str(delay) + " seconds ..")
                    await asyncio.sleep(delay)
                    if delay < MAX_DELAY:
                        delay_increase = (1 + random.random()) * pow(
                            log10(1 + cumulated_deco_time + delay), 2)
                        cumulated_deco_time = cumulated_deco_time + delay
                        delay = min(MAX_DELAY, delay + delay_increase)
            else:
                delay = DELAY_INIT
                cumulated_deco_time = 0
                try:
                    data = await self._recv()
                except websockets.exceptions.WebSocketException as e:
                    log.warning("ws receive failure :" + str(e))
                    await self.disconnect()
                    continue
                msg = json.loads(data)
                log.info("New message {msgtype}".format(**msg))
                log.debug(msg)
                if msg['msgtype'] == "router_config":
                    # Register router_config info
                    keys = ('NetID', 'JoinEui', 'region', 'DRs', 'max_eirp')
                    for k in keys:
                        setattr(self, k, None)
                        # Some LNS (like AWS) do not respect the casing of
                        # the router_config fields. To make sure we do
                        # not miss them, look for non case sensitive matches
                        for msgKey in msg.keys():
                            if k.lower() == msgKey.lower():
                                setattr(self, k, msg.get(msgKey))

                    # Also put msg in queue to forward and
                    # handle radio configuration
                    await self.rxq.put(msg)
                    self.computeDevAddrRanges()

                    self.stats_joineui_filter = []
                    if self.JoinEui is not None:
                        for JOIN_EUI in self.JoinEui:
                            log.info("JoinEui allowed from " +
                                     hex(JOIN_EUI[0]) +
                                     " to " + hex(JOIN_EUI[1]))
                            self.stats_joineui_filter.append(
                                [f'{JOIN_EUI[0]:x}', f'{JOIN_EUI[1]:x}'])
                    self.router_config.set()
                elif msg['msgtype'] == "dnmsg":
                    # Put msg in queue, use recv() to get it
                    await self.rxq.put(msg)
                else:
                    log.warning(
                        "Unhandled message type \
                        \"{msgtype}\"".format(**msg))

    async def recv(self):
        return await self.rxq.get()

    if os.environ.get('INJECTOR_PORT') is not None:
        def inj_log(self, S):
            log.error("-------------------------------")
            log.error(S)
            log.error("-------------------------------")

        async def handle_injection(self, reader, writer):
            data = await reader.read(32*1024)
            message = data.decode()
            addr = writer.get_extra_info('peername')

            self.inj_log(f"Received {message!r} from {addr!r}")
            writer.write(b"OK")
            await writer.drain()

            self.inj_log("Closing the injector connection")
            writer.close()

            self.inj_log(" DOWNLINK INJECTION ")
            J = json.loads(message.replace("'", "\""))
            await self.rxq.put(J)

        async def injector_task(self, port):
            server = await asyncio.start_server(
                self.handle_injection, '127.0.0.1', port)

            addrs = ', '.join(str(sock.getsockname())
                              for sock in server.sockets)
            self.inj_log(f" Injector Serving on {addrs} ")

            async with server:
                await server.serve_forever()

    def get_credentials(self):
        creds = ['tcuri', 'trust', 'crt', 'key']
        return [getattr(self, m, None) for m in creds]
